import openai

class OpenAI:
    def __init__(self, model="gpt-4o-mini", api_key=None):
        self.client = openai.Client(
            api_key=api_key
        )
        self.model = model

    def chat(self, messages, tools, response_format=None):
        return self.client.chat.completions.create(
            messages=messages,
            model=self.model,
            response_format=response_format,
            tools=tools
        )
    
    def chat_parsed(self, messages, tools, response_format=None):
        return self.client.beta.chat.completions.parse(
            messages=messages,
            model=self.model,
            response_format=response_format,
            tools=tools
        ).choices[0].message.parsed


from typing import List
from pydantic import BaseModel

from ..llm import OpenAI
from .chunker import BaseChunker

class Chunks(BaseModel):
    chunks: List[str]
class AgenticChunker(BaseChunker):
    def __init__(self, llm_client):
        self.llm_client = llm_client or OpenAI(model="gpt-4o-mini")
    
    def get_chunks(self, text):
        # To be implemented
        pass


import crochet

from ..chunkers import BasicChunker
crochet.setup()

import uuid
from scrapy.spiders import Rule, CrawlSpider
from scrapy.linkextractors import LinkExtractor
from bs4 import BeautifulSoup
from scrapy import signals
from scrapy.crawler import CrawlerRunner
from scrapy.signalmanager import dispatcher

from ..embedding import OpenAIEmbedClient

class WebSpider(CrawlSpider):
    name = "web-scraper"
    
    crawled_urls = []
    
    def __init__(
        self,
        urls=None,
        follow_links=True,
        restrict_navigation_css=None,
        restrict_css=None,
        ignore_css=None,
        *args,
        **kwargs
    ):
        print("Initializing ContentSpider")
        self.rules = (
            Rule(LinkExtractor(
                restrict_css=restrict_navigation_css,
                unique=True
            ), callback='parse_page', follow=follow_links),
        )
        super(WebSpider, self).__init__(*args, **kwargs)
        self.start_urls = urls
        self.restrict_css = restrict_css
        self.ignore_css = ignore_css
    
    def parse_page(self, response):
        print("Crawling 1:", response.url)
        soup = BeautifulSoup(response.body, 'html.parser')
        try:
            if(self.ignore_css):
                for element in soup.select(selector=self.ignore_css):
                    element.decompose()
            if(self.restrict_css):
                content = " ".join([p.get_text() for p in soup.select(selector=self.restrict_css)])
            else:
                content = soup.get_text()
            print("Content:", content)
            yield {
                "content": content,
                "url": response.url,
                "title": response.css('title::text').get()
            }
        except Exception as e: print(e)
        
        
        
class WebScraper():
    
    MAX_LINK_FOLLOW = 1000
    
    def __init__(
        self,
        urls=None,
        follow_links=True,
        restrict_navigation_css=None,
        restrict_css=None,
        ignore_css=None,
        db_provider=None,
        embedding_client=OpenAIEmbedClient(),
        chunker=BasicChunker(),
        group=None
    ):
        self.crawl_runner = CrawlerRunner()
        self.urls = urls
        self.follow_links = follow_links
        self.restrict_navigation_css = restrict_navigation_css
        self.restrict_css = restrict_css
        self.ignore_css = ignore_css
        self.db_provider = db_provider
        self.chunker = chunker
        self.embedding_client = embedding_client
        self.group = group
        
    @crochet.run_in_reactor
    def crawl(self):
        print("Crawling 2:", self.urls)
        dispatcher.connect(self._crawler_result, signal=signals.item_scraped)
        try:
            self.eventual = self.crawl_runner.crawl(
                WebSpider,
                urls=self.urls,
                follow_links=self.follow_links,
                restrict_navigation_css=self.restrict_navigation_css,
                restrict_css=self.restrict_css,
                ignore_css=self.ignore_css
            )
        except Exception as e: print(e)
        
    def cancel(self):
        self.eventual.cancel()
    
    def _crawler_result(self, item, response, spider):
        print("Crawling 3:", item)
        for chunk in self.chunker.get_chunks(item["content"]):
            print("Crawling 3:", chunk)
            try:
                self.db_provider.insert({
                    "id": str(uuid.uuid1()),
                    "title": item["title"],
                    "content": chunk,
                    "ref": item["url"],
                    "embedding": self.embedding_client.embed(chunk),
                    "type": "web",
                    "group": self.group
                })
            except Exception as e: print(e)



from llama_index.core import Document
from llama_index.core.node_parser import (
    SemanticSplitterNodeParser,
)

from .chunker import BaseChunker
from ..embedding import OpenAIEmbedClient

class SemanticChunker(BaseChunker):
    def __init__(self, buffer_size=1, breakpoint_percentile_threshold=95, api_key=None, embedding_client=None):
        embed_model = embedding_client or OpenAIEmbedClient()
        self.node_parser = SemanticSplitterNodeParser(
            buffer_size=buffer_size, breakpoint_percentile_threshold=breakpoint_percentile_threshold, embed_model=embed_model
        )
    
    def get_chunks(self, text):
        nodes = self.node_parser.get_nodes_from_documents([Document(text=text)])
        chunks = []
        for node in nodes:
            chunks.append(node.get_content())
        return chunks


from llama_index.core import Document
from llama_index.core.node_parser import SentenceSplitter

from .chunker import BaseChunker

class BasicChunker(BaseChunker):
    def __init__(self, chunk_size=500, chunk_overlap=50):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.node_parser = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    
    def get_chunks(self, text):
        nodes = self.node_parser.get_nodes_from_documents([Document(text=text)])
        chunks = []
        for node in nodes:
            chunks.append(node.get_content())
        return chunks


import os
from peewee import MySQLDatabase, Model, CharField, TextField, FloatField, SQL
from tidb_vector.peewee import VectorField

tidb = MySQLDatabase(
    database=os.environ.get('TIDB_DATABASE', 'test'),
    user=os.environ.get('TIDB_USERNAME', 'root'),
    password=os.environ.get('TIDB_PASSWORD', ''),
    host=os.environ.get('TIDB_HOST', 'localhost'),
    port=int(os.environ.get('TIDB_PORT', '4000')),
    ssl_verify_cert = True,
    ssl_verify_identity = True
)

class KB(Model):
    class Meta:
        database = tidb
        table_name = 'kb'
    
    id = CharField(primary_key=True)
    title = TextField()
    content = TextField()
    ref = TextField()
    type = TextField()
    group = TextField()
    embedding = VectorField(1536, constraints=[SQL("COMMENT 'hnsw(distance=cosine)'")])

class TiDBProvider():
    def connect(self):
        tidb.connect()
        if(not tidb.table_exists('kb')):
            tidb.create_tables([KB])
            
    def get_client(self):
        return tidb
            
    def disconnect(self):
        tidb.close()
        
    def insert(self, kb):
        try:
            KB.create(
                id=kb['id'],
                title=kb['title'],
                content=kb['content'],
                ref=kb['ref'],
                embedding=kb['embedding'],
                type=kb['type'],
                group=kb['group']
            )
        except Exception as e: print(e)
        
    def search(self, vector, group=None, limit = 5):
        distance = KB.embedding.cosine_distance(vector).alias('distance')
        results = KB.select(KB.id, KB.title, KB.content, KB.ref, KB.type, distance).where(group == group).order_by(distance).limit(limit)
        return list(results.dicts())


import json

class KBBot:

    tools = [
        {
            "type": "function",
            "function": {
                "name": "search_kb",
                "description": "Search for articles in the knowledge base. The search will happen with a vector representation of the query.",
                "parameters": {
                    "type": "object",
                    "properties":{
                        "query": {
                            "type": "string",
                            "description": "The query to search for. The vector representation of the query will be used to search for similar articles in the knowledge base."
                        }
                    },
                    "required": ["query"]
                },
            }
        }
    ]    

    def __init__(
        self,
        db_provider,
        embedding_client,
        llm_client,
        group=None,
        tasks_prompt="",
        history=[]
    ):
        self.db_provider = db_provider
        self.embedding_client = embedding_client
        self.llm_client = llm_client
        self.group = group
        self.tasks_prompt = tasks_prompt
        self.history = history
        
    def system_prompt(self):
        return f"""
        You are a assistant helping people to find information in a knowledge base. You have access to a database of knowledge base articles.
        Tasks:
        - Search for articles in the knowledge base.
        {self.tasks_prompt}
        Important:
        - You can search the articles by typing 'search_kb'.
        - You will answer only the questions related to the knowledge base articles.
        - If you don't know the answer, politely ask ofr more information if needed, or say you don't know.
        - Always give references to the articles you provide information from.
        """
        
    def search_kb(
        self,
        query
    ):
        vector = self.embedding_client.embed(query)
        results = self.db_provider.search(vector, group=self.group)
        return results
    
    def chat(
        self,
        message
    ):
        messages = [
            {
                "role": "system",
                "content": self.system_prompt()
            },
            *self.history,
            {
                "role": "user",
                "content": message
            }
        ]
        response = self.llm_client.chat(messages=messages, tools=self.tools)
        if(response.choices[0].message.tool_calls and len(response.choices[0].message.tool_calls) > 0):
            tool_call_responses = []
            for tool_call in response.choices[0].message.tool_calls:
                kbs = self.search_kb(json.loads(tool_call.function.arguments)["query"])
                tool_call_responses.append({
                    "role": "tool",
                    "content": json.dumps({
                        "kbs": kbs
                    }),
                    "tool_call_id": tool_call.id
                })
            response = self.llm_client.chat(messages=[*messages, response.choices[0].message, *tool_call_responses], tools=self.tools)
        return response.choices[0].message.content